/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.6 $ IRCAM $Date: 1993/11/19 16:22:06 $
 *
 * FTS by Miller Puckette
 *
 */

/*
	This version of the pitch tracker is intended for voice but might later
	be generalized to solve other score following problems.
*/


#include "m_extern.h"
#include "d_graph.h"
#include "d_ugen.h"
#define __STRICT_BSD__
#include "math.h"
#undef __STRICT_BSD__

double sigpitch_mtof(double f)
{
  return (440. * exp(.057762265 * (f - 69.)));
}

double sigpitch_ftom(double f)
{
  if (f > 0) return (69. + (1./.057762265) * log(f/440.));
  else return (0);
}

float sigpitch_getfloat(t_atom *at)
{
  float f;
  if (at->a_type == A_LONG) 
    f = at->a_w.w_long;
  else if (at->a_type == A_FLOAT)
    f = at->a_w.w_float;
  else if (sscanf(at->a_w.w_sym, "%f", &f) < 1) {
    warn("%s: bad number", at->a_w.w_sym->s_name);
    f = 0;
  }
  return f;
}

sigpitch_ilog2(int n)
{
  int ret = -1;
  while (n)
    {
      n >>= 1;
      ret++;
    }
  return (ret);
}

/******************* PITCHTRACK *********************/
#define NAMPCOEF 8
#define MINPOINTS 128
#define MAXPOINTS 512
#define MAXLOG 9
#define FMAXPOINTS 512.

#define PEROCTAVE 24
#define FPEROCTAVE 24.
#define PERHALFTONE 2

#define NKERNEL (PEROCTAVE * MAXLOG - 2 * PEROCTAVE)
#define FREQSTEP 1.0293022
#define LOGSTEP .028881096
#define PI 3.14159265
#define TWOPI (2. * PI)
#define CUTOFF .05
#define ONQUALITYTHRESH .5
#define OFFQUALITYTHRESH .25
#define ONPOWTHRESH .01
#define OFFPOWTHRESH .005
#define STICK .3
#define OVERLAP 2.

#define EXTRAKERNEL 73		/* number of extra filter kernels to keep at end */
#define ALIGN 16			/* byte alignment for microcode */
#define LOWINDEX 6
#define DEFREATTACK .5
#define DEFRATIME 200
#define DEFDEBOUNCETIME 150

#define DUMTAB1SIZE 256
#define DUMTAB2SIZE 1024
static float dumbtab1[DUMTAB1SIZE], dumbtab2[DUMTAB2SIZE];
#define HISTORY 16
#define POWCHUNK 64

typedef struct kernel
{
  float *k_coef;
  int k_ncoef;
  int k_onset;
  float k_powcorrect;
} t_kernel;

typedef struct sigpitch
{
  t_head x_h;
  t_sig *x_io[IN1+OUT0];
  void *x_clock;
  void *x_qualityoutlet;
  void *x_wesseloutlet;
  void *x_evenoutlet;
  void *x_poweroutlet;
  void *x_micropitchoutlet;
  void *x_attackoutlet;
  void *x_pitchoutlet;

  float x_quality;
  float x_wessel;
  float x_even;
  float x_power;
  float x_micropitch;
  float x_newpitch;
  int x_attack;

  float *x_buf;
  float *x_ebuf;
  float *x_fillbuf;
  int x_npts;
  int x_nchan;
  float x_lofreq;
  float x_reffreq;		/* frequency corresponding to MIDI 69 */
  float x_stretchcoef;	/* inverse natural log of a MIDI halftone */
  float x_onpowthresh;
  float x_offpowthresh;
  float x_onqualitythresh;
  float x_offqualitythresh;
  float x_stick;
  float x_ampcoef[NAMPCOEF];
  long x_glisstime;
  float x_halftonesperframe;
  float x_srate;
  int x_powtime;
  int x_powbuckets;
  float x_powslopethresh;
  float x_lastpow;
  int x_lopitch;
  int x_hipitch;
  int x_lopoint;
  int x_hipoint;
  int x_lastindex;
  int x_midipit;		/* current pitch */
  int x_overlap;
  float *x_freeme;
  int x_printme;
  int x_peaked;
  long x_midilastout;
  long x_miditime;
  long x_debouncetime;
  float x_maxerror;
  int x_histindex;
  float x_pitchhist[HISTORY];
  float x_amphist[HISTORY];
  int x_buckets;
  float x_vibrato;
  long x_vibtime;
  int x_countup;
} t_sigpitch;

int sigpitch_doit (t_sigpitch *x);
void sigpitch_fixpitch(t_sigpitch *x);
void sigpitch_glisstime(t_sigpitch *x, long int n);
void sigpitch_vibrato(t_sigpitch *x, long vibtime, double vibdepth);
void sigpitch_reattack(t_sigpitch *x, double f, long int n);

float sigpitch_defcoef[] =
{ 1., .9, .8, .7, .7, .6, .6, .6};

t_kernel *pt_kernel;
float *sigpitch_wind256, *sigpitch_wind512;
extern int h_ishost;

double dumsqrt(double d)
{
  float f = d;
  long l = *(long *)(&f);
  if (f < 0) return (0.);
  else return (dumbtab1[(l >> 23) & 0xff] * dumbtab2[(l >> 13) & 0x3ff]);
}

pt__init()
{
  float *specspace, *espec;
  /* estimate total size of kernels: cover the whole spectrum 
     4 times and overlap 4 points apiece */
  int specsize = 4 * NKERNEL + OVERLAP * MAXPOINTS, filtno;
  int i;
  float *fftbuf, filtfreq, filtomega;
  t_kernel *kp;
  char fftspace[2*sizeof(float) * MAXPOINTS + UCALL_PAGE_SIZE], *foo;
  if (pt_kernel) return;
#ifdef FTS
  if (!(specspace = (float *)getbytes(specsize * sizeof(*specspace))))
    {
      error("pitchtrack: can't allocate space");
      return;
    }
  espec = specspace + specsize;
  if (!(pt_kernel = (t_kernel *)getbytes(NKERNEL * sizeof(*pt_kernel))))
    {
      freebytes(specspace, (specsize * sizeof(*specspace)));
      error("pitchtrack: can't allocate space");
      return;
    }
  if (!(foo = (char *)getbytes(SIGALIGN+(512+256)*sizeof(float))))
    {
      freebytes(specspace, (specsize * sizeof(*specspace)));
      freebytes(pt_kernel, NKERNEL * sizeof(*pt_kernel));
      error("pitchtrack: can't allocate space");
      return;
    }
  sigpitch_wind256 = (float *)(foo + ((-(long)foo) & (SIGALIGN-1)));
  sigpitch_wind512 = sigpitch_wind256 + 256;
    /* make 256-point and 512-point hanning windows for power calc */
  for (i = 0; i < 256; i++) sigpitch_wind256[i] =
    (1./256.) * (1. - cos((2 * 3.14159/256.) * i));
  for (i = 0; i < 512; i++) sigpitch_wind512[i] =
    (1./512.) * (1. - cos((2 * 3.14159/512.) * i));
  /* get a page-aligned buffer for doing FFT */
  fftbuf = (float *)(((long)fftspace) +
		     ((-(long)fftspace) & (UCALL_PAGE_SIZE - 1)));
  
  for (i = 0, filtfreq = 2. / FMAXPOINTS, kp = pt_kernel; i < NKERNEL; 
       i++, filtfreq *= FREQSTEP, kp++)
    {
      float protowind = 1. / (OVERLAP * (FREQSTEP - 1.) * filtfreq),
      window, phase, wphase,
      filtomega = filtfreq * TWOPI, windomega, cutval, norm,
      midipit;
      float *fp1, *fp2;
      int nbin;
      int j;
      window = (protowind > .5 * FMAXPOINTS ? .5 * FMAXPOINTS : protowind);
      windomega = PI / window;
      cu_zero(fftbuf, 2 * MAXPOINTS);
      norm = 1./window;
      for (fp1 = fp2 = fftbuf + MAXPOINTS, j = window,
	   phase = 0., wphase = 0.; j--;
	   fp1 -= 2, fp2 += 2, phase += filtomega, wphase += windomega)
	{
	  double wind = norm * (.5 + .5 * cos(wphase));
	  fp1[0] = fp2[0] =  cos(phase) * wind;
	  fp1[1] = -(fp2[1] =  sin(phase) * wind);
	  /* if (!i && j > window-3)
	     {
	     post("1 = %f", (double)1.);
	     post("window %f", window);
	     post("1./window %f", 1./window);
	     post("norm  %f", norm);
	     post("norm > .0001 %d", norm > .0001);
	     post("windomega %f", windomega);
	     } */
	}
      /* if (i == 0)
	 {
	 float *fp3;
	 int z = 10;
	 for (z = 10, fp3 = fftbuf + MAXPOINTS + 2; z -- > 0; fp3 += 2)
	 post("%f %f", fp3[0], fp3[1]);
	 post("at end:");
	 for (z = 10, fp3 = fftbuf + (2 * MAXPOINTS - 20);
	 z -- > 0; fp3 += 2)
	 post("%f %f", fp3[0], fp3[1]);
	 } */
      
      fft(fftbuf, MAXPOINTS, 0);
      /* if (i == 0)
	 {
	 float *fp3;
	 int z;
	 post("---------");
	 for (z = 10, fp3 = fftbuf; z -- > 0; fp3 += 2)
	 post("%f %f", fp3[0], fp3[1]);
	 } */
      cutval = CUTOFF;
      for (j = 0, fp2 = fftbuf + MAXPOINTS; j < (MAXPOINTS/2); j++, fp2 -= 2)
	if (*fp2 >= cutval || *fp2 < (-cutval)) break;
      for (j = 0, fp1 = fftbuf; j < (MAXPOINTS/2); j++, fp1 += 2)
	if (*fp1 >= cutval || *fp1 < (-cutval)) break;
      if (fp1 >= fp2)
	{
	  static int bugged = 0;
	  if (!bugged)
	    {
	      post("bug: pitchtrack %d: zero-length kernel", i);
	      bugged = 1;
	    }
	  fp1 = fp2;
	}
      nbin = 1 + ((fp2 - fp1) >> 1);
      /* if (i == 100 || i == 101)
	 {
	 float *fp3;
	 int z = 10;
	 post("n = %d; start at %d", nbin, (fp1 - fftbuf) >> 1);
	 for (z = 10, fp3 = fp1; fp3 <= fp2 && (z -- > 0); fp3 += 2)
	 post("%f %f", fp3[0], fp3[1]);
	 post("(from zero:");
	 for (z = 10, fp3 = fftbuf; z -- > 0; fp3 += 2)
	 post("%f %f", fp3[0], fp3[1]);
	 } */
      if (specspace + nbin >= espec)
	{
	  post("bug: pitchtrack: misestimated the number of bins");
	  return;
	}
      
      kp->k_onset = fp1 - fftbuf;
      kp->k_coef = specspace;
      kp->k_ncoef = nbin;
      kp->k_powcorrect = window / protowind;
      /* post("%d\n", nbin); */
      for (; nbin--; fp1 += 2) *specspace++ = *fp1;
    }
  for (i = 0; i < DUMTAB1SIZE; i++)
    {
      float f;
      long l = (i ? (i == DUMTAB1SIZE-1 ? DUMTAB1SIZE-2 : i) : 1)<< 23;
      *(long *)(&f) = l;
      dumbtab1[i] = sqrt(f);	
    }
  for (i = 0; i < DUMTAB2SIZE; i++)
    {
      float f = 1 + (1./DUMTAB2SIZE) * i;
      dumbtab2[i] = sqrt(f);	
    }
#endif
}

#ifndef I860
void sigpitch__makespect(float *fftbuf, t_kernel *kp, float *spect, int n)
{
  while (n--)
    {
      float rsum = 0, isum = 0;
      float *fp = fftbuf + kp->k_onset;
      float *cp = kp->k_coef;
      int k = kp->k_ncoef;
      while (k--)
	{
	  float coef = *cp++;
	  rsum += *fp++ * coef;
	  isum += *fp++ * coef;
	}
      *spect++ = rsum * rsum + isum * isum;
      kp++;
    }
}
#else I860
void sigpitch__makespect();
#endif I860

void cu_pt(t_sigpitch *x, float *sig, long int n)
{
  float *fp = x->x_fillbuf;
#ifdef FTS
  sigpitch_copy(sig, fp, n << 2);
  fp += n;
  if (fp != x->x_ebuf) x->x_fillbuf = fp;
  else sigpitch_doit(x);
#endif
}

sigpitch_doit(t_sigpitch *x)
{
#ifdef FTS
  int npts = x->x_npts;
  int nchan = x->x_nchan, index = -1, i, stable,
    lo = x->x_lopoint-2, hi = x->x_hipoint;
  float bestamp, pitchpow, totpow, mspow, newpitch, *fp, *fp2, *fp3, *fftbuf;
  t_kernel *kp = pt_kernel;
  char fftspace[sizeof(float) * (2*MAXPOINTS) + UCALL_PAGE_SIZE];
  float para[3], correct, spectrum[NKERNEL + EXTRAKERNEL],
  pitches[NKERNEL], ampspect[NKERNEL + EXTRAKERNEL];
    /* get a page-aligned buffer for doing FFT */
  fftbuf = (float *)(((long)fftspace) +
		     ((-(long)fftspace) & (UCALL_PAGE_SIZE - 1)));
  fp3 = fftbuf + POWCHUNK;
  cu_zero(fftbuf, 2 * POWCHUNK);	/* clear and warm up cache */
  
    /* get mean-square power */
  for (fp = (x->x_npts == 512 ? sigpitch_wind512 : sigpitch_wind256),
    fp2 = x->x_buf, i = npts; i -= POWCHUNK;
      fp += POWCHUNK, fp2 += POWCHUNK)
  {
    cu_vmul(fp2, fp2, fp3, POWCHUNK);
    cu_vmul(fp, fp3, fp3, POWCHUNK);
    cu_vadd(fp3, fftbuf, fftbuf, POWCHUNK);
  }
  for(i = POWCHUNK, fp = fftbuf, mspow=0; i--;) mspow += *fp++;
  x->x_power = 100 + 8.6858 *log(sqrt(mspow) + .000001);

  cu_zero(fftbuf, 2 * npts);		/* zero it out */
  sigpitch_mux(x->x_buf, fftbuf, 2, npts);	/* copy input into real part */
  x->x_fillbuf = x->x_buf + x->x_overlap;
  if (x->x_overlap) sigpitch_copy(x->x_buf + (npts - x->x_overlap), x->x_buf,
			    x->x_overlap * sizeof(float));
  
  fft(fftbuf, npts, 0);
  /* get a bounded-Q power spectrum, quarter-tone spaced */
  sigpitch__makespect(fftbuf, pt_kernel, spectrum, x->x_nchan);
  for (i = EXTRAKERNEL, fp = spectrum + NKERNEL; i--;) *fp++ = 0;
  
  if (x->x_printme >= 2)
    {
      int z;
      float *fp3 = spectrum;
      post("raw filter output:");
      for (z = 0; z < nchan; z += 4, fp3 += 4)
	post("%3.3d %12.1f %12.1f %12.1f %12.1f",
	     z, fp3[0], fp3[1], fp3[2], fp3[3]);
    }
  
  /* make an amplitude spectrum */
  for (i = NKERNEL, fp = spectrum, fp2 = ampspect; i--;)
#ifdef I860
    *fp2++ = dumsqrt(*fp++);
#else I860
  *fp2++ = sqrt(*fp++);
#endif I860
  for (i = EXTRAKERNEL; i--;) *fp2++ = 0;
  
  
  /* look for bin with max. power in (itself and first few
     overtones.) */
  for (i = 0, fp = ampspect, fp2 = x->x_ampcoef, bestamp=0;
       i < nchan; fp++, i++)
    {
      float mofo = fp[0] * fp2[0] +
	fp[24] * fp2[1] +
	  fp[38] * fp2[2] +
	    fp[48] * fp2[3] +
	      fp[56] * fp2[4] +
		fp[62] * fp2[5] +
		  fp[67] * fp2[6] +
		    fp[72] * fp2[7];

      pitches[i] = mofo;
      if (i >= lo && i < hi && mofo > bestamp) bestamp = mofo, index = i;
    }
  if (index <= 0)			/* didn't find anything at all? */
  {
    newpitch = 0;
    x->x_lastindex = 0;
    x->x_quality = x->x_wessel = x->x_even = 0;
    goto done;
  }
  
  if (x->x_lastindex)		/* possibly stick to previous pitch */
  {
    int lastindex = x->x_lastindex;
    float x2;
    fp = pitches + lastindex;
    x2 = *fp;
    if (fp[-1] > x2 && lastindex > lo) x2 = *--fp, lastindex--;
    else if (fp[1] > x2 && lastindex < hi - 1) x2 = *++fp, lastindex++;
    if (x2 > (1. - x->x_stick) * bestamp) index = lastindex;
  }
  
  /* get total power for the good overtones */
  fp = spectrum + index;
  pitchpow = fp[0] + fp[24] + fp[38] + fp[48] + fp[56] +
    fp[62] + fp[67] + fp[72];
  
  /* get the total power from DC up to the 8th harmonic.
     The k_powcorrect factor is to correct for the fact that the
     filterbank overlap is greater in the low bands and so
     merely summing them would weight them too highly. */
  fp += 72;
  if (fp >= spectrum + NKERNEL) fp = spectrum + (NKERNEL - 1);
  for (fp2 = spectrum, totpow = 0., kp = pt_kernel; fp2 < fp; fp2++, kp++)
    totpow += *fp2 * kp->k_powcorrect;
  
  if (x->x_printme)
    post("index %d, value %f; overall %f", index,
	 (float)sqrt(pitchpow)/npts, (float)sqrt(totpow)/npts);
  
  x->x_quality = pitchpow / totpow;

  /* to report a pitch there must be at least "offqualitythresh"
     times the total power in the good bins. */
  if (pitchpow < x->x_offpowthresh ||
      x->x_quality < x->x_offqualitythresh)
  {
    newpitch = 0;
    x->x_lastindex = 0;
    x->x_quality = x->x_wessel = x->x_even = 0;
    if (x->x_printme) post("no pitch (q or power too low)");
    goto done;
  }

  fp = spectrum + index;
  x->x_wessel = (fp[0] + 2. * fp[24] + 3. * fp[38] + 4. * fp[48] +
    4. * fp[56] + 6. * fp[62] + 7. * fp[67] + 8. * fp[72]) / pitchpow;
  x->x_even = (fp[24] + fp[48] + fp[62] + fp[72]) / pitchpow;

  /* remember the index for pitch-stick test next time */
  if (pitchpow > x->x_onpowthresh) x->x_lastindex = index;
  else x->x_lastindex = 0;
  
  /* do a parabolic correction to increase precision */
  for (i = 0, fp = spectrum + index - 1; i <= 3; i++, fp++)
    para[i] = fp[0] + 2.*fp[24] + 3.*fp[38] + 4.*fp[48] +
      5.*fp[56] + 6.*fp[62] + 7.*fp[67] + 8.*fp[72];
  if (para[0] >= para[1]) correct = -.5;
  else if (para[2] >= para[1]) correct = .5;
  else
  {
    double f1 = 1. - para[0]/para[1], f2 = 1. - para[2]/para[1];
    correct = (f1-f2)/(2.*(f1+f2) - 3.*f1*f2);
  }
  /* correct = .5 * (para[2] - para[0])
     / (2. * para[1] - para[0] - para[2]); */
  
  newpitch = 69. + (12./log(2.)) * (
  	log(x->x_lofreq/x->x_reffreq) +
  	x->x_stretchcoef * LOGSTEP * (index + correct));
  stable = (x->x_glisstime == 0 ||
       (newpitch >= x->x_micropitch - x->x_halftonesperframe &&
	newpitch <= x->x_micropitch + x->x_halftonesperframe));
  if (x->x_printme)
  {
    post("pitch %f, %sstable", newpitch, (stable ? "" : "un"));
  }
done:
  x->x_histindex = (x->x_histindex + 1)&(HISTORY-1);
  x->x_pitchhist[x->x_histindex] = newpitch;
  x->x_amphist[x->x_histindex] = mspow;
  if (x->x_midipit)
  {
    float error;
    /* test for turnoff -- if "newpitch" is more than 3/4
	of a halftone away from the current note, (or from an
	octave up or down from it) the current note has stopped
	playing.  This also takes care of the case where
	newpitch == 0. */
    error = newpitch - x->x_midipit;
    if (error > 6) error -= 12;
    else if (error < -6) error += 12;
    if (error > x->x_vibrato || error < -x->x_vibrato)
    {
      x->x_midipit = 0;
    }
   }
    /* now test for reattack, which is a condition on the
    	amplitude envelope.  Reattack is also marked by setting current
	pitch to zero; the note-on conditions must apply before a new
	note is output. */
  if (!x->x_peaked)
  {
    if (x->x_powbuckets &&
      (mspow * (x->x_powslopethresh + 1)
	< x->x_amphist[(x->x_histindex - x->x_powbuckets) & (HISTORY-1)]))
	    x->x_peaked = 1;
  }
  else
  {
    if (x->x_powbuckets && pitchpow > x->x_onpowthresh
      && (mspow 
	> (x->x_powslopethresh + 1) *
	  x->x_amphist[(x->x_histindex - x->x_powbuckets) & (HISTORY-1)]))
    {
      x->x_attack = 1;
      /* post("---- %f %f", pitchpow, x->x_onpowthresh); */
      x->x_peaked = 0;
	/* here's the good part -- if there's a reattack and we haven't
	reported a note yet since the last one, look again RETROACTIVELY,
	using laxer criteria */
      if (x->x_countup < 100)
      {
	float bestpow = 0, bestpitch, recentpow = mspow;
	int histdepth = (x->x_countup > HISTORY ? HISTORY : x->x_countup),
	  gobackup = 0;
	for (i = 1; i < histdepth; i++)
	{
	  float npow = x->x_amphist[(x->x_histindex - i) & (HISTORY-1)];
	  if (!gobackup)
	  {
	    if (npow > recentpow)
	      gobackup = 1;
	    recentpow = npow;
	  }
	  else if (npow > bestpow)
	  {
	    bestpow = npow;
	    bestpitch = x->x_pitchhist[(x->x_histindex - i) & (HISTORY-1)];
	  }
	}
	post("----   %d   %f   %f", x->x_countup, bestpow, bestpitch);
	if (bestpow > 0) x->x_newpitch = bestpitch;
      }
      x->x_midipit = 0;
      x->x_countup = 0;
    }
  }
  x->x_countup++;
  x->x_lastpow = mspow;
  
  /* test for turnon.  We must: not have a current pitch;
     have a candidate pitch; have at least a certain
     amount of power and "quality". */
  if (!x->x_midipit && newpitch != 0 &&
      pitchpow > x->x_onpowthresh &&
      x->x_quality > x->x_onqualitythresh)
  {
    int ipit, deviated = 0;
    float error, pitsum;
    if (x->x_printme) post("error %f", error);
      /* in addition we must have a stable pitch in the sense that it
      does not deviate more than x->x_vibrato from the mean over the
      last x->x_buckets values. */
	
    ipit = newpitch + .5;
    error = newpitch - ipit;
    if (x->x_buckets)
    {
      for (pitsum = 0, i = 0; i < x->x_buckets; i++)
    	pitsum += x->x_pitchhist[(x->x_histindex-i)&(HISTORY-1)];
      pitsum /= x->x_buckets;
      for (i = 0; i < x->x_buckets; i++)
      {
    	float deviation =
	  x->x_pitchhist[(x->x_histindex-i)&(HISTORY-1)] - pitsum;
	if (x->x_printme) post("deviation %f", deviation);

	if (deviation < -x->x_vibrato || deviation > x->x_vibrato)
	{
	  if (x->x_printme) post("called off");
	  deviated = 1;
	}
      }
    }
    
    if ((error < x->x_maxerror) && (error > -x->x_maxerror) && !deviated)
    {
      int rats = x->x_midilastout;
      if (gettime() > x->x_miditime ||
	  (ipit != rats && ipit != rats + 12 && ipit != rats - 12))
      {
	x->x_miditime = gettime() + x->x_debouncetime;
	x->x_midilastout = ipit;
	x->x_peaked = 0;
	x->x_midipit = pitsum+.5;
	x->x_newpitch = pitsum;
	x->x_countup = 1000;
      }
    }
    else if (x->x_printme) post("out of tune");
  }
  x->x_printme = 0;
  x->x_micropitch = newpitch;
  clock_delay(x->x_clock, 0L);
#endif FTS
}

void sigpitch_put(t_sigpitch *x, long int whether)
{
  if (whether)
    {
      u_stdout((t_ugen *)x);
      x->x_srate = x->x_io[0]->s_sr;
      sigpitch_fixpitch(x);
      sigpitch_glisstime(x, x->x_glisstime);
      sigpitch_vibrato(x, x->x_vibtime, x->x_vibrato);
      sigpitch_reattack(x, x->x_powslopethresh, x->x_powtime);
      dspchain_addc(cu_pt, 3,
		    x, x->x_io[0]->s_shit, x->x_io[0]->s_n);
    }
}

void sigpitch_bang(t_sigpitch *x)
{
}

void sigpitch_pt_tune(t_sigpitch *x, double f)
{
  if (f >= 1.) x->x_reffreq = sigpitch_mtof(f);
  else x->x_reffreq = 440.;
}

void sigpitch_stretch(t_sigpitch *x, double f)
{
  if (f < .01) f = 1;
  x->x_stretchcoef = f;
}

void sigpitch_ampl_threshold(t_sigpitch *x, double f)
{
  post("warning: ampl_threshold: does nothing; use amp-range");
}

void sigpitch_quality_threshold(t_sigpitch *x, double f)
{
  post("warning: quality_threshold is obsolete; use quality-range");
  x->x_onqualitythresh = f;
  x->x_offqualitythresh = .5*f;
}

void sigpitch_amp_range(t_sigpitch *x, double f1, double f2)
{
  double scale = x->x_npts * x->x_npts;
  x->x_offpowthresh = f1 * f1 * scale;
  x->x_onpowthresh = f2 * f2 * scale;
}

void sigpitch_quality_range(t_sigpitch *x, double f1, double f2)
{
  x->x_offqualitythresh = f1 * f1;
  x->x_onqualitythresh = f2 * f2;
}

void sigpitch_pitch_range(t_sigpitch *x, long int n1, long int n2)
{
  x->x_lopitch = n1;
  x->x_hipitch = n2;
  sigpitch_fixpitch(x);
}

void sigpitch_fixpitch(t_sigpitch *x)
{
  double pitchoffirst =
		sigpitch_ftom(x->x_lofreq = 2. * x->x_srate / x->x_npts);
  int offset1 = .5 + 2 * (x->x_lopitch - pitchoffirst);
  int offset2 = .5 + 2 * (x->x_hipitch - pitchoffirst);
  if (offset1 < LOWINDEX) offset1 = LOWINDEX;
  if (offset1 >= x->x_nchan) offset1 = x->x_nchan-1;
  if (offset2 < offset1) offset2 = offset1;
  if (offset2 >= x->x_nchan) offset2 = x->x_nchan-1;
  x->x_lopoint = offset1;
  x->x_hipoint = offset2;
}

void sigpitch_stick(t_sigpitch *x, double f)
{
  if (f < 0.) f = 0.;
  else if (f > 1.) f = 1.;
  x->x_stick = f;
}

void sigpitch_maxerror(t_sigpitch *x, double f)
{
  if (f < 0.) f = 0.;
  else if (f > .5) f = .5;
  x->x_maxerror = f;
}

void sigpitch_coefs(t_sigpitch *x, t_symbol *s, int ac, t_atom *av)
{
  if (ac < NAMPCOEF) error("sigpitch_coefs; need %d values", NAMPCOEF);
  else
    {
      int i;
      for (i = 0; i < NAMPCOEF; i++)
	{
	  float f = sigpitch_getfloat(av+i);
	  if (f < 0) f = 0;
	  else if (f > 1) f = 1;
	  x->x_ampcoef[i] = f;
	}
    }
}

void sigpitch_glisstime(t_sigpitch *x, long int n)
{
  if (n > 0)
    {
      x->x_glisstime = n;
      x->x_halftonesperframe =
	(x->x_npts - x->x_overlap)/(.001 * n * x->x_srate);
    }
  else x->x_glisstime = 0;
}

void sigpitch_vibrato(t_sigpitch *x, long vibtime, double vibdepth)
{
  if (vibdepth < 0) vibdepth = 0;
  if (vibtime < 0) vibtime = 0;
  x->x_vibtime = vibtime;
  x->x_buckets =
    (vibtime * .001 * x->x_srate)/(x->x_npts - x->x_overlap);
  if (x->x_buckets >= HISTORY) x->x_buckets = HISTORY-1;
  x->x_vibrato = vibdepth;
}

void sigpitch_debouncetime(t_sigpitch *x, long int n)
{
  x->x_debouncetime = (n > 0 ? n : 0);
}

void sigpitch_reattack(t_sigpitch *x, double f, long int n)
{
  if (f > 0 && n > 0)
  {
    x->x_powslopethresh = f;
    x->x_powtime = (n < 1 ? 1 : n);
    x->x_powbuckets =
      (n * .001 * x->x_srate)/(x->x_npts - x->x_overlap);
    if (x->x_powbuckets >= HISTORY) x->x_powbuckets = HISTORY-1;
  }
  else
  {
    x->x_powslopethresh = 0;
    x->x_powtime = 0;
    x->x_powbuckets = 0;
  }
}

void sigpitch_print(t_sigpitch *x, long int n)
{
  if (n < 1) n = 1;
  x->x_printme = n;
  post("pitch tracker:");
  post("reference-pitch %f", sigpitch_ftom(x->x_reffreq));
  post("stretch %f", x->x_stretchcoef);
  post("stick %f", x->x_stick);
  post("gliss-time %d", x->x_glisstime);
  post("max-error %f", x->x_maxerror);
  post("debounce-time %d", x->x_debouncetime);
  post("reattack %f %d", x->x_powslopethresh, x->x_powtime);
  post("coefs %f %f %f %f",
       x->x_ampcoef[0],x->x_ampcoef[1],x->x_ampcoef[2],x->x_ampcoef[3]); 
  post("...   %f %f %f %f",
       x->x_ampcoef[4],x->x_ampcoef[5],x->x_ampcoef[6],x->x_ampcoef[7]); 
  post("amp-range %f %f",
       sqrt(x->x_offpowthresh)/x->x_npts,
       sqrt(x->x_onpowthresh)/x->x_npts);
  post("quality-range %f %f", sqrt(x->x_offqualitythresh),
       sqrt(x->x_onqualitythresh));
  post("pitch-range %d %d", x->x_lopitch, x->x_hipitch);
  if (x->x_lofreq > 1.)
    {
      post("(possible pitch range: %d - %d)",
	   (int)(sigpitch_ftom(x->x_lofreq) + (.5 + LOWINDEX/2)),
	   (int)(sigpitch_ftom(x->x_lofreq) + (.5 + x->x_nchan/2)));
    }
}

void sigpitch_tick(t_sigpitch *x)
{
  outlet_float(x->x_wesseloutlet, x->x_wessel);
  outlet_float(x->x_evenoutlet, x->x_even);
  outlet_float(x->x_qualityoutlet, x->x_quality);
  outlet_float(x->x_poweroutlet, x->x_power);
  outlet_float(x->x_micropitchoutlet, x->x_micropitch);
  if (x->x_attack)
  {
    outlet_bang(x->x_attackoutlet);
    x->x_attack = 0;
  } 
  if (x->x_newpitch > 0)
  {
    outlet_float(x->x_pitchoutlet, (double)(x->x_newpitch));
    x->x_newpitch = 0;
  }
}

void sigpitch_clean(t_sigpitch *x)
{
  clock_free(x->x_clock);
  freebytes(x->x_freeme, x->x_npts * sizeof(float) + ALIGN);
  u_clean((t_ugen *)x);
}

t_externclass *sigpitch_class;

void *sigpitch_new(long int npoints, int period)
{
  t_sigpitch *x;
  float *fp;
  int logn, i;
  if (!npoints) npoints = 512;
  logn = sigpitch_ilog2(npoints);
  if (npoints != (1 << logn))
    {
      error("pt: power of two required as argument");
      return (0);
    }
  if (npoints > MAXPOINTS || npoints < MINPOINTS)
    {
      error("pt: argument out of range");
      return (0);
    }
  if (period & (MAXVS-1))
    {
      error("pt: period must be a multiple of %d", MAXVS);
      return (0);
    }
  if (period > npoints)
    {
      error("pt: period can't exceed window size");
      return (0);
    }
  if (!period) period = npoints;
  if (!(fp = (float *)
	getbytes(npoints * sizeof(float) + ALIGN)))
    {
      error("pt: out of memory");
      return (0);
    }
  x = (t_sigpitch *)obj_new((t_messlist **)&sigpitch_class, (void *)0);
  
  x->x_freeme = fp;
  fp = (float *)(((-(long)fp) & (ALIGN - 1)) + (long)fp);
  cu_zero(fp, npoints);
  u_setup((t_ugen *)x, IN1, OUT0);

  x->x_wesseloutlet = outlet_new(x, gensym("float"));
  x->x_evenoutlet = outlet_new(x, gensym("float"));
  x->x_qualityoutlet = outlet_new(x, gensym("float"));
  x->x_poweroutlet = outlet_new(x, gensym("float"));
  x->x_micropitchoutlet = outlet_new(x, gensym("float"));
  x->x_attackoutlet = outlet_new(x, gensym("bang"));
  x->x_pitchoutlet = outlet_new(x, gensym("float"));

  x->x_clock = clock_new(x, sigpitch_tick);
  fft_init();
  pt__init();
  
  x->x_srate = 44100.;
  x->x_overlap = npoints-period;
  x->x_buf = x->x_fillbuf = fp;
  x->x_ebuf = x->x_buf + npoints;
  x->x_npts = npoints;
  x->x_nchan = PEROCTAVE * logn - 2 * PEROCTAVE;
  x->x_lofreq = 1.;
  x->x_printme = 0;
  x->x_reffreq = 440.;
  x->x_maxerror = .4;
  sigpitch_stretch(x, 1.);
  sigpitch_amp_range(x, OFFPOWTHRESH, ONPOWTHRESH);
  sigpitch_quality_range(x, OFFQUALITYTHRESH, ONQUALITYTHRESH);
  sigpitch_pitch_range(x, 0, 1000);
  sigpitch_reattack(x, DEFREATTACK, DEFRATIME);
  sigpitch_debouncetime(x, DEFDEBOUNCETIME);
  x->x_midipit = 0;
  x->x_midilastout = x->x_miditime = 0;
  x->x_lastindex = 0;
  x->x_stick = STICK;
  x->x_peaked = 0;
  x->x_histindex = 0;
  x->x_newpitch = 0;
  x->x_countup = 1000;
  x->x_attack = 0;
  x->x_quality = x->x_wessel = x->x_even = x->x_micropitch = 0;
  for(i = 0; i < HISTORY; i++) x->x_pitchhist[i] = x->x_amphist[i] = 0;
  for (i = 0; i < NAMPCOEF; i++) x->x_ampcoef[i] = sigpitch_defcoef[i];
  sigpitch_glisstime(x, 0);
  sigpitch_vibrato(x, 0, .75);
  return (x);
}

void sigpitch_setup()
{
  c_extern(&sigpitch_class, sigpitch_new, sigpitch_clean, gensym("pitch~"),
		sizeof(t_sigpitch), 0, A_DEFLONG, A_DEFLONG, 0);
  c_addmess(sigpitch_put, gensym("put"), A_CANT, 0);
  c_addbang(sigpitch_bang);
  c_addmess(sigpitch_pt_tune, gensym("pt_tune"), A_FLOAT, 0);
  c_addmess(sigpitch_ampl_threshold, gensym("ampl_threshold"), A_FLOAT, 0);
  c_addmess(sigpitch_quality_threshold, gensym("quality_threshold"),
	    A_FLOAT, 0);
  c_addmess(sigpitch_pt_tune, gensym("reference-pitch"), A_FLOAT, 0);
  c_addmess(sigpitch_maxerror, gensym("max-error"), A_FLOAT, 0);
  c_addmess(sigpitch_stretch, gensym("stretch"), A_FLOAT, 0);
  c_addmess(sigpitch_amp_range, gensym("amp-range"), A_FLOAT, A_FLOAT, 0);
  c_addmess(sigpitch_quality_range, gensym("quality-range"),
	    A_FLOAT, A_FLOAT, 0);
  c_addmess(sigpitch_pitch_range, gensym("pitch-range"), A_LONG, A_LONG, 0);
  c_addmess(sigpitch_stick, gensym("stick"), A_FLOAT, 0);
  c_addmess(sigpitch_coefs, gensym("coefs"), A_GIMME, 0);
  c_addmess(sigpitch_glisstime, gensym("gliss-time"), A_LONG, 0);
  c_addmess(sigpitch_vibrato, gensym("vibrato"), A_LONG, A_FLOAT, 0);
  c_addmess(sigpitch_debouncetime, gensym("debounce-time"), A_LONG, 0);
  c_addmess(sigpitch_reattack, gensym("reattack"), A_FLOAT, A_LONG, 0);
  c_addmess(sigpitch_print, gensym("print"), A_DEFLONG, 0);
  u_inletmethod(0);
}

